/*
 * Main coding by Elliot Lee <sopwith@redhat.com>, Red Hat Software. 
 * Copyright (C) 1996.
 * Copyright (c) Jan Rêkorajski, 1999.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <syslog.h>
#include <errno.h>
#include <pwd.h>

#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

#include "md5.h"
#include "pam_localpwcheck.h"

/*
 * i64c - convert an integer to a radix 64 character
 */
static int i64c(int i)
{
	if (i < 0)
		return ('.');
	else if (i > 63)
		return ('z');
	if (i == 0)
		return ('.');
	if (i == 1)
		return ('/');
	if (i >= 2 && i <= 11)
		return ('0' - 2 + i);
	if (i >= 12 && i <= 37)
		return ('A' - 12 + i);
	if (i >= 38 && i <= 63)
		return ('a' - 38 + i);
	return ('\0');
}

char *crypt_md5_wrapper(const char *pass_new, size_t pass_new_len)
{
	/*
	 * Code lifted from Marek Michalkiewicz's shadow suite. (CG)
	 * removed use of static variables (AGM)
	 */

	struct timeval tv;
	MD5_CTX ctx;
	unsigned char result[16];
	char *cp = (char *) result;
	unsigned char tmp[16];
	int i;
	char *x = NULL;

	GoodMD5Init(&ctx);
	gettimeofday(&tv, (struct timezone *) 0);
	GoodMD5Update(&ctx, (void *) &tv, sizeof tv);
	i = getpid();
	GoodMD5Update(&ctx, (void *) &i, sizeof i);
	i = clock();
	GoodMD5Update(&ctx, (void *) &i, sizeof i);
	GoodMD5Update(&ctx, result, sizeof result);
	GoodMD5Final(tmp, &ctx);
	strcpy(cp, "$1$");	/* magic for the MD5 */
	cp += strlen(cp);
	for (i = 0; i < 8; i++)
		*cp++ = i64c(tmp[i] & 077);
	*cp = '\0';

	/* no longer need cleartext */
	x = Goodcrypt_md5_ex(pass_new, pass_new_len, (const char *) result);

	return x;
}

int check_old_password(const char *forwho, const char *newpass, size_t newpasslen)
{
	static char buf[16384];
	char *s_luser, *s_uid, *s_npas, *s_pas;
	int retval = PAM_SUCCESS;
	FILE *opwfile;
	size_t user_len;

	opwfile = fopen(_PATH_OPASSWD, "r");
	if (opwfile == NULL) {
		return PAM_IGNORE;
	}

	user_len = strlen(forwho);

	while (fgets(buf, 16380, opwfile)) {
		if (!strncmp(buf, forwho, user_len) &&
		    buf[user_len] == ':') {
			buf[strlen(buf) - 1] = '\0';
			s_luser = strtok(buf, ":,");
			s_uid = strtok(NULL, ":,");
			s_npas = strtok(NULL, ":,");
			s_pas = strtok(NULL, ":,");
			while (s_pas != NULL) {
				char *md5pass = Goodcrypt_md5_ex(newpass, newpasslen, s_pas);
				if (!strcmp(md5pass, s_pas)) {
					_pam_overwrite(md5pass);
					_pam_drop(md5pass);
					retval = PAM_AUTHTOK_ERR;
					break;
				}
				s_pas = strtok(NULL, ":,");
				_pam_overwrite(md5pass);
				_pam_drop(md5pass);
			}
			break;
		}
	}
	fclose(opwfile);

	return retval;
}

int save_old_password(const char *forwho, const char *oldpass,
		      size_t oldpasslen, int howmany)
{
	static char buf[16384];
	static char nbuf[16384];
	char *s_luser, *s_uid, *s_npas, *s_pas, *pass;
	int npas;
	FILE *pwfile, *opwfile;
	int err = 0;
	int oldmask;
	int found = 0;
	struct passwd *pwd = NULL;

	if (howmany < 0) {
		return PAM_SUCCESS;
	}

	if (oldpass == NULL) {
		return PAM_SUCCESS;
	}

	oldmask = umask(077);
	pwfile = fopen(_PATH_NOPASSWD, "w");
	umask(oldmask);
	if (pwfile == NULL) {
		return PAM_AUTHTOK_ERR;
	}

	opwfile = fopen(_PATH_OPASSWD, "r");
	if (opwfile == NULL) {
		syslog(LOG_INFO, "pam_localpwcheck: %s does not exist, creating", _PATH_OPASSWD);
	}

	chown(_PATH_NOPASSWD, 0, 0);
	chmod(_PATH_NOPASSWD, 0600);

	if (opwfile != NULL) {
		while (fgets(buf, 16380, opwfile)) {
			if (!strncmp(buf, forwho, strlen(forwho))) {
				buf[strlen(buf) - 1] = '\0';
				s_luser = strtok(buf, ":");
				s_uid = strtok(NULL, ":");
				s_npas = strtok(NULL, ":");
				s_pas = strtok(NULL, ":");
				npas = strtol(s_npas, NULL, 10) + 1;
				while (npas > howmany) {
					s_pas = strpbrk(s_pas, ",");
					if (s_pas != NULL)
						s_pas++;
					npas--;
				}
				pass = crypt_md5_wrapper(oldpass, oldpasslen);
				if (s_pas == NULL)
					snprintf(nbuf, sizeof(nbuf), "%s:%s:%d:%s\n",
							s_luser, s_uid, npas, pass);
				else
					snprintf(nbuf, sizeof(nbuf), "%s:%s:%d:%s,%s\n",
							s_luser, s_uid, npas, s_pas, pass);
				_pam_overwrite(pass);
				_pam_drop(pass);
				if (fputs(nbuf, pwfile) < 0) {
					err = 1;
					break;
				}
				found = 1;
			} else if (fputs(buf, pwfile) < 0) {
				err = 1;
				break;
			}
		}
		fclose(opwfile);
	}
	if (!found) {
		pwd = getpwnam(forwho);
		if (pwd == NULL) {
			err = 1;
		} else {
			pass = crypt_md5_wrapper(oldpass, oldpasslen);
			snprintf(nbuf, sizeof(nbuf), "%s:%d:1:%s\n",
					forwho, pwd->pw_uid, pass);
			_pam_overwrite(pass);
			_pam_drop(pass);
			if (fputs(nbuf, pwfile) < 0) {
				err = 1;
			}
		}
	}

	if (fclose(pwfile)) {
		syslog(LOG_INFO, "pam_localpwcheck: error writing entries to old passwords file: %s",
			strerror(errno));
		err = 1;
	}

	if (!err) {
		rename(_PATH_NOPASSWD, _PATH_OPASSWD);
		return PAM_SUCCESS;
	} else {
		unlink(_PATH_NOPASSWD);
		return PAM_AUTHTOK_ERR;
	}
}
