/*
 * Copyright (c) 2004-2006 PADL Software Pty Ltd.
 * All rights reserved.
 * Use is subject to license.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <syslog.h>
#include <errno.h>
#include <pwd.h>
#include <ctype.h>
#ifdef HAVE_SHADOW_H
#include <shadow.h>
#endif
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

#include "pam_localpwcheck.h"

/*
 * A short-and-sweet password strength checking module for
 * local accounts.
 * 
 * Highly unconfigurable (this is supposed to be a feature!)
 */

#define STRLENOF(_x)	(sizeof(_x) - 1)

static void lpw_debug(int lpw_flags, const char *fmt, ...)
{
	va_list args;

	if ((lpw_flags & LPW_FLAGS_DEBUG) == 0) {
		return;
	}

	va_start(args, fmt);
	vsyslog(LOG_DEBUG, fmt, args);
	va_end(args);
}

/* Return 1 if callback() returns a non-zero value for any character in the password */
static int lpw_match_against_class(size_t passlen,
	const char *password,
	int (*callback)(int))
{
	int i, found = 0;

	for (i = 0; i < passlen; i++) {
		if ((*callback)(password[i]) != 0) {
			found++;
			break;
		}
	}

	i = 0;

	return found;
}

+/* Return 1 if callback() returns a zero-value for any character in the password */
static int lpw_match_against_class_negative(size_t passlen,
	const char *password,
	int (*callback)(int))
{
	int i, found = 0;

	for (i = 0; i < passlen; i++) {
		if (lpw_match_against_class(1, &password[i], callback) == 0) {
			found++;
			break;
		}
	}

	return (found > 0);
}

static int lpw_count_classes(const char *password, size_t passlen)
{
	int n_classes = 0;

	n_classes += lpw_match_against_class(passlen, password, isupper);
	n_classes += lpw_match_against_class(passlen, password, islower);
	n_classes += lpw_match_against_class(passlen, password, isdigit);
	n_classes += lpw_match_against_class_negative(passlen, password, isalnum);

	return n_classes;
}

#if defined(HAVE_SHADOW_H) && (defined(HAVE_FGETSPENT) || defined(HAVE_FGETSPENT_R))
static int lpw_getspnam_r(const char *user,
	struct spwd *result_buf,
	char *buffer,
	size_t buflen,
	struct spwd **result)
{
	FILE *fp;
	int nerrno = 0;
	int oerrno = errno;
	struct spwd *sp = NULL;

	*result = NULL;

	errno = 0;

	fp = fopen(_PATH_SHADOW, "r");
	if (fp == NULL) {
		nerrno = errno;
		goto out;
	}

	while (1) {
#if FGETSPENT_R_ARGS == 5
		nerrno = fgetspent_r(fp, result_buf, buffer, buflen, &sp);
		if (nerrno != 0)
			break;
#elif FGETSPENT_R_ARGS == 4
		sp = fgetspent_r(fp, result_buf, buffer, buflen);
		if (sp == NULL)
			break;
#else
		sp = fgetspent(fp);
		if (sp == NULL)
			break;
#endif

		if (strcmp(sp->sp_namp, user) == 0) {
			*result = sp;
			break;
		}
	}

out:
	if (fp != NULL) {
		if (sp == NULL && nerrno == 0) {
			nerrno = ENOENT;
		}
		fclose(fp);
	}

	errno = oerrno;
	return nerrno;
}
#endif

#if defined(HAVE_FGETPWENT) || defined(HAVE_FGETPWENT_R)
static int lpw_getpwnam_r(const char *user,
	struct passwd *result_buf,
	char *buffer,
	size_t buflen,
	struct passwd **result)
{
	FILE *fp;
	int nerrno = 0;
	int oerrno = errno;
	struct passwd *pw = NULL;

	*result = NULL;

	errno = 0;

	fp = fopen(_PATH_PASSWD, "r");
	if (fp == NULL) {
		nerrno = errno;
		goto out;
	}

	while (1) {
#if FGETPWENT_R_ARGS == 5
		nerrno = fgetpwent_r(fp, result_buf, buffer, buflen, &pw);
		if (nerrno != 0)
			break;
#elif FGETPWENT_R_ARGS == 4
		pw = fgetpwent_r(fp, result_buf, buffer, buflen);
		if (pw == NULL)
			break;
#else
		pw = fgetpwent(fp);
		if (pw == NULL)
			break;
#endif

		if (strcmp(pw->pw_name, user) == 0) {
			*result = pw;
			break;
		}
	}

out:
	if (fp != NULL) {
		if (pw == NULL && nerrno == 0) {
			nerrno = ENOENT;
		}
		fclose(fp);
	}

	errno = oerrno;
	return nerrno;
}
#else
#error fgetpwent() or fgetpwent_r() must be present
#endif

static int lpw_get_authtok(pam_handle_t *pamh,
	int flags,
	const char *prompt,
	char **p)
{
	int rc;
	struct pam_message msg[1], *pmsg[1];
	struct pam_response *resp;
	struct pam_conv *conv;

	*p = NULL;

	pmsg[0] = &msg[0];
	msg[0].msg_style = PAM_PROMPT_ECHO_OFF;
	msg[0].msg = (CONST_ARG char *)prompt;
	resp = NULL;

	rc = pam_get_item(pamh, PAM_CONV, (CONST_ARG void **)&conv);
	if (rc != PAM_SUCCESS) {
		return rc;
	}

	rc = (conv->conv)(1, (CONST_ARG struct pam_message **)pmsg,
			  &resp, conv->appdata_ptr);
	if (rc != PAM_SUCCESS) {
		return rc;
	}

	if (resp != NULL) {
		if ((flags & PAM_DISALLOW_NULL_AUTHTOK) && 
		    resp[0].resp == NULL) {
			free(resp);
			return PAM_AUTH_ERR;
		}
	
		*p = resp[0].resp;
		/* leak if resp[0].resp is malloced. */
		resp[0].resp = NULL;
	} else {
#ifdef PAM_AUTHTOK_RECOVERY_ERR
		return PAM_AUTHTOK_RECOVERY_ERR;
#else
		return PAM_AUTHTOK_RECOVER_ERR;
#endif /* PAM_AUTHTOK_RECOVERY_ERR */
	}

	free(resp);

	return PAM_SUCCESS;
}

static int lpw_authenticate(pam_handle_t *pamh,
	int flags,
	struct passwd *pw,
	const char *oldauthtok)
{
#ifdef HAVE_SHADOW_H
	struct spwd *sp = NULL, sp_buf;
	char buf[BUFSIZ];
#endif
	int rc;

	if (oldauthtok == NULL) {
		return PAM_AUTH_ERR;
	}

	rc = PAM_AUTH_ERR;

#ifdef HAVE_SHADOW_H
	if (strcmp(pw->pw_passwd, "x") == 0) {
		(void) lpw_getspnam_r(pw->pw_name, &sp_buf, buf, sizeof(buf), &sp);
		if (sp == NULL) {
			return PAM_AUTH_ERR;
		}

		if (strcmp(crypt(oldauthtok, sp->sp_pwdp), sp->sp_pwdp) == 0) {
			rc = PAM_SUCCESS;
		}
	} else
#endif /* HAVE_SHADOW_H */
	if (strcmp(crypt(oldauthtok, pw->pw_passwd), pw->pw_passwd) == 0) {
		rc = PAM_SUCCESS;
	}

	return rc;
}

static int lpw_send_message(pam_handle_t *pamh,
	const char *message,
	int style)
{
	struct pam_message msg, *pmsg;
	struct pam_response *resp;
	struct pam_conv *conv;
	int rc;

	pmsg = &msg;

	rc = pam_get_item(pamh, PAM_CONV, (CONST_ARG void **)&conv);
	if (rc != PAM_SUCCESS) {
		return rc;
	}

	msg.msg_style = style;
	msg.msg = (char *)message;
	resp = NULL;

	return (conv->conv)(1, (CONST_ARG struct pam_message **) &pmsg,
			    &resp, conv->appdata_ptr);
}

static int lpw_authtok_check_history(pam_handle_t *pamh,
	const char *authtok,
	size_t authtoklen,
	lpw_policy_t *policy)
{
	int rc;
	char *user;

	rc = pam_get_item(pamh, PAM_USER, (CONST_ARG void **)&user);
	if (rc != PAM_SUCCESS) {
		lpw_debug(policy->flags, "pam_localpwcheck: pam_get_item(PAM_USER) failed: %s",
			pam_strerror(pamh, rc));
		return rc;
	}

	rc = check_old_password(user, authtok, authtoklen);
	if (rc != PAM_SUCCESS && rc != PAM_IGNORE) {
		lpw_debug(policy->flags, "pam_localpwcheck: password history check failed: %s",
			pam_strerror(pamh, rc));
		return rc;
	}

	return PAM_SUCCESS;
}

static int lpw_check_authtok(pam_handle_t *pamh,
	const char *oldauthtok,
	const char *authtok,
	lpw_policy_t *policy)
{
	int classes;
	size_t authtoklen;
	char buf[256];
	const char *msg = NULL;
	int rc;

	if (authtok == NULL) {
		return PAM_AUTHTOK_ERR;
	}

	if (oldauthtok != NULL && strcmp(oldauthtok, authtok) == 0) {
		msg = "Passwords must differ";
		goto fail;
	}

	authtoklen = strlen(authtok);

	if (authtoklen < policy->min_length) {
		snprintf(buf, sizeof(buf), "Password too short - must be at least %d characters.", policy->min_length);
		msg = buf;
		goto fail;
	}

	if (policy->check_length != -1 && authtoklen > policy->check_length) {
		authtoklen = policy->check_length; /* clamp effective length for checking */
	}

	classes = lpw_count_classes(authtok, authtoklen);
	if (classes < policy->min_classes) {
		snprintf(buf, sizeof(buf), "Password must contain characters, numerals, or punctuation.");
		msg = buf;
		goto fail;
	}

	rc = lpw_authtok_check_history(pamh, authtok, authtoklen, policy); /* XXX authtoklen ignored */
	if (rc != PAM_SUCCESS && rc != PAM_IGNORE) {
		snprintf(buf, sizeof(buf), "Password cannot repeat a previously used password.");
		msg = buf;
		goto fail;
	}

	return PAM_SUCCESS;

fail:
	rc = lpw_send_message(pamh, msg, PAM_ERROR_MSG);
	if (rc != PAM_SUCCESS) {
		return rc;
	}

	return PAM_AUTHTOK_ERR;
}

static void lpw_cleanup_policy(pam_handle_t *pamh,
	void *data,
	int error_status)
{
	if (data != NULL) {
		_pam_drop(data);
	}
}

static int lpw_get_policy(pam_handle_t *pamh,
	int flags,
	int argc,
	const char **argv,
	lpw_policy_t **ppolicy)
{
	int i;
	int rc;
	lpw_policy_t *policy;

	*ppolicy = NULL;

	policy = (lpw_policy_t *)calloc(1, sizeof(*policy));
	if (policy == NULL) {
		return PAM_BUF_ERR;
	}

	policy->flags = 0;
	policy->min_length = LPW_MIN_LENGTH;
	policy->min_classes = LPW_MIN_CLASSES;
	policy->max_tries = LPW_MAX_TRIES;
	policy->history_length = LPW_HISTORY_LENGTH;
	policy->check_length = LPW_CHECK_LENGTH;

	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i], "debug") == 0)
			policy->flags |= LPW_FLAGS_DEBUG;
		else if (strcmp(argv[i], "use_authtok") == 0)
			policy->flags |= LPW_FLAGS_USE_AUTHTOK;
		else if (strcmp(argv[i], "ask_oldauthtok") == 0)
			policy->flags |= LPW_FLAGS_GET_OLDAUTHTOK;
		else if (strcmp(argv[i], "check_oldauthtok") == 0)
			policy->flags |= LPW_FLAGS_VALIDATE_OLDAUTHTOK;
		else if (strncmp(argv[i], "minlen=", STRLENOF("minlen=")) == 0)
			policy->min_length = atoi(argv[i] + STRLENOF("minlen="));
		else if (strncmp(argv[i], "minclasses=", STRLENOF("minclasses=")) == 0)
			policy->min_classes = atoi(argv[i] + STRLENOF("minclasses="));
		else if (strncmp(argv[i], "retry=", STRLENOF("retry=")) == 0)
			policy->max_tries = atoi(argv[i] + STRLENOF("retry="));
		else if (strncmp(argv[i], "remember=", STRLENOF("remember=")) == 0)
			policy->history_length = atoi(argv[i] + STRLENOF("remember="));
		else if (strncmp(argv[i], "checklen=", STRLENOF("checklen=")) == 0)
			policy->check_length = atoi(argv[i] + STRLENOF("checklen="));
		else
			syslog(LOG_ERR, "pam_localpwcheck: illegal option %s", argv[i]);
	}

	rc = pam_set_data(pamh, PADL_LPW_POLICY_DATA, policy, lpw_cleanup_policy);
	if (rc != PAM_SUCCESS) {
		lpw_cleanup_policy(pamh, policy, rc);
		return rc;
	}

	*ppolicy = policy;

	return PAM_SUCCESS;
}

static int lpw_ignore_user(const char *user,
	int argc,
	const char **argv)
{
	int i;
	const char *ignore = NULL;

	for (i = 0; i < argc; i++) {
		if (strncmp(argv[i], "ignore=", STRLENOF("ignore=")) == 0) {
			ignore = argv[i] + STRLENOF("ignore=");
			break;
		}
	}

	if (ignore == NULL || strcmp(ignore, user) != 0) {
		return PAM_SUCCESS;
	}

	return PAM_IGNORE;
}

static int lpw_isprivileged(pam_handle_t *pamh)
{
	char *service;
	int rc;

	rc = pam_get_item(pamh, PAM_SERVICE, (CONST_ARG void **)&service);
	if (rc == PAM_SUCCESS && strcmp(service, "passwd") == 0) {
		return (getuid() == 0);
	}

	return 0;
}

static int lpw_check_update(pam_handle_t *pamh,
	int flags,
	int argc,
	const char **argv)
{
	int rc;
	lpw_policy_t *policy = NULL;
	char *user;
	char *authtok;

	rc = pam_get_data(pamh, PADL_LPW_POLICY_DATA, (const void **)&policy);
	if (rc == PAM_NO_MODULE_DATA) {
		/* not a local user; PRELIM_CHECK should have set this for local users */
		return PAM_IGNORE;
	} else if (rc != PAM_SUCCESS) {
		return rc;
	}

	if (policy == NULL) {
		/* XXX should not get here */
		return PAM_IGNORE;
	}

	rc = pam_get_item(pamh, PAM_USER, (CONST_ARG void **)&user);
	if (rc != PAM_SUCCESS) {
		lpw_debug(policy->flags, "pam_localpwcheck: pam_get_item(PAM_USER) failed: %s",
			pam_strerror(pamh, rc));
		return rc;
	}

	lpw_debug(policy->flags, "pam_localpwcheck: UPDATE_AUTHTOK called for user %s "
		"(min_length=%d min_classes=%d max_tries=%d flags=%04x)",
		user, policy->min_length, policy->min_classes, policy->max_tries, policy->flags);

	rc = pam_get_item(pamh, PAM_AUTHTOK, (CONST_ARG void **)&authtok);
	if (rc != PAM_SUCCESS) {
		lpw_debug(policy->flags, "pam_localpwcheck: failed to get new authentication token: %s",
			pam_strerror(pamh, rc));
		return rc;
	} else if (authtok == NULL) {
		lpw_debug(policy->flags, "pam_localpwcheck: new authentication token is NULL");
		return PAM_AUTHTOK_ERR;
	}

	/* Write into history */
	if (policy->history_length != -1) {
		size_t authtoklen = strlen(authtok);

		if (policy->check_length != -1 && authtoklen > policy->check_length) {
			authtoklen = policy->check_length; /* clamp effective length for checking */
		}

		rc = save_old_password(user, authtok, authtoklen, policy->history_length);
		if (rc == PAM_SUCCESS) {
			lpw_debug(policy->flags, "pam_localpwcheck: wrote password history for user");
		} else {
			lpw_debug(policy->flags, "pam_localpwcheck: failed to write password history: %s",
				pam_strerror(pamh, rc));
		}
	} else {
		/*
		 * We should really return PAM_IGNORE but it doesn't appear to
		 * be handled correctly on Linux-PAM.
		 *
		 * See:
		 * http://www.informit.com/articles/article.asp?p=30602&seqNum=3&rl=1
		 */
		rc = PAM_SUCCESS;
	}

	return rc;
}

static int lpw_check_prelim(pam_handle_t *pamh,
	int flags,
	int argc,
	const char **argv)
{
	int rc;
	char *authtok = NULL;
	const char *user = NULL;
	const char *oldauthtok = NULL;
	int tries = 0;
	lpw_policy_t *policy;
	struct passwd *pw, pw_buf;
	char buf[BUFSIZ];
#ifdef PAM_REPOSITORY
	pam_repository_t rep_data;
#endif

	rc = pam_get_item(pamh, PAM_USER, (CONST_ARG void **)&user);
	if (rc != PAM_SUCCESS) {
		syslog(LOG_INFO, "pam_localpwcheck: pam_get_item(PAM_USER) failed: %s",
			pam_strerror(pamh, rc));
		return rc;
	}

	rc = lpw_ignore_user(user, argc, argv);
	if (rc != PAM_SUCCESS) {
		return rc;
	}

	rc = lpw_getpwnam_r(user, &pw_buf, buf, sizeof(buf), &pw);
	if (rc != 0 || pw == NULL) {
		/* Not a local user */
		return PAM_IGNORE;
	}

	rc = lpw_get_policy(pamh, flags, argc, argv, &policy);
	if (rc != PAM_SUCCESS) {
		return rc;
	}

#ifdef PAM_REPOSITORY
	/* force pam_unix to change through files, just to be sure */
	rep_data.type = "files";
	rep_data.scope = NULL;
	rep_data.scope_len = 0;

	rc = pam_set_item(pamh, PAM_REPOSITORY, &rep_data);
	if (rc != PAM_SUCCESS) {
		lpw_debug(policy->flags, "pam_localpwcheck: pam_set_item(PAM_REPOSITORY) failed: %s",
			pam_strerror(pamh, rc));
		return rc;
	}
#endif /* PAM_REPOSITORY */

	lpw_debug(policy->flags, "pam_localpwcheck: PRELIM_CHECK called for user %s "
		"(min_length=%d min_classes=%d max_tries=%d flags=%04x)",
		user, policy->min_length, policy->min_classes, policy->max_tries, policy->flags);

	/*
	 * On Solaris this will be set by the pam_authtok_get module. There is
	 * no such module on Linux so we will need to get it ourselves if we
	 * weren't stacked after another password changing module.
	 */
	rc = pam_get_item(pamh, PAM_OLDAUTHTOK, (CONST_ARG void **)&oldauthtok);
	if (rc != PAM_SUCCESS) {
		lpw_debug(policy->flags, "pam_localpwcheck: pam_get_item(PAM_OLDAUTHTOK) failed: %s",
			pam_strerror(pamh, rc));
		return rc;
	}

	/*
	 * NB: this is really just to make Linux-PAM work. Fundamentally
	 * it seems a broken design that the chauthtok() module is
	 * required to authenticate the user's existing password. The
	 * current Solaris approach is to have a passwd_auth module
	 * that does this in the *authentication* phase.
	 *
	 * lpw_isprivileged() tests for PAM_SERVICE == passwd and the
	 * actual UID == 0.
	 */
	if (oldauthtok == NULL && lpw_isprivileged(pamh) == 0) {
		char *tmpauthtok;

		if (policy->flags & LPW_FLAGS_GET_OLDAUTHTOK) {
			rc = lpw_get_authtok(pamh, flags, OLD_PASSWORD_PROMPT, &tmpauthtok);
			if (rc != PAM_SUCCESS) {
				lpw_debug(policy->flags, "pam_localpwcheck: failed to get old authentication token: %s",
					pam_strerror(pamh, rc));
				return rc;
			}

			rc = pam_set_item(pamh, PAM_OLDAUTHTOK, tmpauthtok);
			if (rc != PAM_SUCCESS) {
				lpw_debug(policy->flags, "pam_localpwcheck: pam_set_item(PAM_OLDAUTHTOK) failed: %s",
					pam_strerror(pamh, rc));
				_pam_overwrite(tmpauthtok);
				_pam_drop(tmpauthtok);
				return rc;
			}

			_pam_overwrite(tmpauthtok);
			_pam_drop(tmpauthtok);

			rc = pam_get_item(pamh, PAM_OLDAUTHTOK, (CONST_ARG void **)&oldauthtok);
			if (rc != PAM_SUCCESS) {
				lpw_debug(policy->flags, "pam_localpwcheck: pam_get_item(PAM_OLDAUTHTOK) failed: %s",
					pam_strerror(pamh, rc));
				return rc;
			}
		}

		if (policy->flags & LPW_FLAGS_VALIDATE_OLDAUTHTOK) {
			rc = lpw_authenticate(pamh, flags, pw, oldauthtok);
			if (rc != PAM_SUCCESS) {
				lpw_debug(policy->flags, "pam_localpwcheck: failed to authenticate against passwd file: %s",
					pam_strerror(pamh, rc));
				return rc;
			} else {
				lpw_debug(policy->flags, "pam_localpwcheck: authenticated user %s against passwd file",
					pw->pw_name);
			}
		}
	}

	if (policy->flags & LPW_FLAGS_USE_AUTHTOK) {
		rc = pam_get_item(pamh, PAM_AUTHTOK, (CONST_ARG void **)&authtok);
		if (rc != PAM_SUCCESS) {
			lpw_debug(policy->flags, "pam_localpwcheck: pam_get_item(PAM_AUTHTOK) failed: %s",
				pam_strerror(pamh, rc));
			return rc;
		}

		rc = lpw_check_authtok(pamh, oldauthtok, authtok, policy);
		if (rc != PAM_SUCCESS) {
			lpw_debug(policy->flags, "pam_localpwcheck: failed to check new authentication token: %s",
				pam_strerror(pamh, rc));
			return rc;
		}

		return PAM_SUCCESS;
	}

	for (tries = 0; tries < policy->max_tries; tries++) {
		char *vfyauthtok = NULL;

		rc = lpw_get_authtok(pamh, flags, NEW_PASSWORD_PROMPT, &authtok);
		if (rc != PAM_SUCCESS) {
			lpw_debug(policy->flags, "pam_localpwcheck: failed to get new authentication token: %s",
				pam_strerror(pamh, rc));
			break;
		}

		rc = lpw_check_authtok(pamh, oldauthtok, authtok, policy);
		if (rc != PAM_SUCCESS) {
			lpw_debug(policy->flags, "pam_localpwcheck: failed to check new authentication token: %s",
				pam_strerror(pamh, rc));
			continue;
		}

		rc = lpw_get_authtok(pamh, flags, AGAIN_PASSWORD_PROMPT, &vfyauthtok);
		if (rc != PAM_SUCCESS) {
			lpw_debug(policy->flags, "pam_localpwcheck: failed to verify authentication token: %s",
				pam_strerror(pamh, rc));
			break;
		}

		if (strcmp(authtok, vfyauthtok) == 0) {
			rc = PAM_SUCCESS;
			break;
		}

		(void) lpw_send_message(pamh, "You must enter the same password", PAM_ERROR_MSG);
		lpw_debug(policy->flags, "pam_localpwcheck: new/verify authentication tokens differ");
		_pam_overwrite(vfyauthtok);
		_pam_drop(vfyauthtok);
		rc = PAM_MAXTRIES;
	}

	if (rc == PAM_SUCCESS) {
		rc = pam_set_item(pamh, PAM_AUTHTOK, authtok);
		if (rc != PAM_SUCCESS) {
			lpw_debug(policy->flags, "pam_localpwcheck: pam_set_item(PAM_AUTHTOK) failed: %s",
				pam_strerror(pamh, rc));
			return rc;
		}
	} else {
		_pam_overwrite(authtok);
	}

	_pam_drop(authtok);

	lpw_debug(policy->flags, "pam_localpwcheck: finished for user %s: returning %d/%s",
		user, rc, pam_strerror(pamh, rc));

	return rc;
}

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh,
	int flags,
	int argc,
	const char **argv)
{
	if (flags & PAM_PRELIM_CHECK) {
		return lpw_check_prelim(pamh, flags, argc, argv);
	} else if (flags & PAM_UPDATE_AUTHTOK) {
		return lpw_check_update(pamh, flags, argc, argv);
	}

	syslog(LOG_ERR, "pam_localpwcheck: called with unknown flags %08x", flags);

	return PAM_SERVICE_ERR;
}
